
__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import logging
import pprint
import threading

import radical.utils as ru

from ... import utils     as rpu
from ... import states    as rps
from ... import constants as rpc


# ------------------------------------------------------------------------------
#
# 'enum' for RPs's pilot scheduler types
#
SCHEDULER_NAME_CONTINUOUS      = "CONTINUOUS"
SCHEDULER_NAME_CONTINUOUS_FIFO = "CONTINUOUS_FIFO"
SCHEDULER_NAME_SCATTERED       = "SCATTERED"
SCHEDULER_NAME_HOMBRE          = "HOMBRE"
SCHEDULER_NAME_TORUS           = "TORUS"
SCHEDULER_NAME_YARN            = "YARN"
SCHEDULER_NAME_SPARK           = "SPARK"


# ------------------------------------------------------------------------------
#
# An RP agent scheduler will place incoming units onto a set of cores and gpus.
#
# This is the agent scheduler base class.  It provides the framework for
# implementing diverse scheduling algorithms and mechanisms, tailored toward
# specific workload types, resource configurations, batch systems etc.
#
# The base class provides the following functionality to the implementations:
#
#   - obtain configuration settings from config files and environments
#   - create aself._nodes list to represent available resources; 
#   - general control and data flow:
#
#       # main loop
#       self._handle_unit(unit):  # unit arrives
#         try_allocation(unit)   # placement is attempted
#         if success:
#            advance(unit)       # pass unit to executor
#         else:
#            wait.append(unit)   # place unit in a wait list
#
#   - notification management:
#     - the scheduler receives notifications about units which completed
#       execution, and whose resources can now be used again for other units,
#     - the above triggers an 'unschedule' (free resources) action and also a
#       `schedule` action (check waitlist if waiting units can now be placed).
#
#
# A scheduler implementation will derive from this base class, and overload the
# following three methods:
#
#   _configure():
#     - make sure that the base class configuration is usable
#     - do any additional configuration
#
#   _allocate_slot(cud):
#     - given a unit description, find and return a suitable allocation
#
#   _release_slot(slots):
#     - release the given allocation
#
#
# The scheduler needs (in the general case) three pieces of information:
#
#   - the layout of the resource (nodes, cores, gpus);
#   - the current state of those (what cores/gpus are used by other units)
#   - the requirements of the unit (single/multi node, cores, gpus)
#
# The first part (layout) is provided by the LRMS, in the form of a nodelist:
#
#    nodelist = [{name : 'node_1', cores: 16, gpus : 2},
#                {name : 'node_2', cores: 16, gpus : 2},
#                ...
#               ]
#
# That is then mapped into an internal representation, which is really the same
# but allows to keep track of resource usage, by setting the fields to
# `rpc.FREE == '-'` or `rpc.BUSY == '#'`:
#
#    nodelist = [{name : 'node_1', cores: [----------------], gpus : [--]},
#                {name : 'node_2', cores: {----------------], gpus : [--]},
#                ...
#               ]
#
# When allocating a set of resource for a unit (2 cores, 1 gpu), we can now
# record those as used:
#
#    nodelist = [{name : 'node_1', cores: [##--------------], gpus : [#-]},
#                {name : 'node_2', cores: {----------------], gpus : [--]},
#                ...
#               ]
#
# This solves the second part from our list above.  The third part, unit
# requirements, are obtained from the unit dict passed for scheduling: the unit
# description contains requests for `cores` and `gpus`.
#
# Note that the unit description will also list the number of processes and
# threads for cores and gpus, and also flags the use of `mpi`, 'openmp', etc.
# The scheduler will have to make sure that for each process to be placed, the
# given number of additional cores are available *and reserved* to create
# threads on.  The threads are created by the application.  Note though that
# this implies that the launcher must pin cores very carefully, to not constrain
# the thread creation for the application.
#
# The scheduler algorithm will then attempt to find a suitable set of cores and
# gpus in the nodelist into which the unit can be placed.  It will mark those as
# `rpc.BUSY`, and attach the set of cores/gpus to the unit dictionary, as (here
# for system with 8 cores & 1 gpu per node):
#
#     unit = { ...
#       'cpu_processes'   : 4,
#       'cpu_process_type': 'mpi',
#       'cpu_threads'     : 2,
#       'gpu_processes    : 2,
#       'slots' :
#       {                 # [[node,   node_uid,   [cpu idx],        [gpu idx]]]
#         'nodes'         : [[node_1, node_uid_1, [[0, 2], [4, 6]], [[0]    ]],
#                            [node_2, node_uid_2, [[1, 3], [5, 7]], [[0]    ]]],
#         'cores_per_node': 8,
#         'gpus_per_node' : 1,
#         'lm_info'       : { ... }
#       }
#     }
#
# The `cpu idx` field is a list of sets, where in each set the first core is
# where an application process is places, while the other cores are reserved for
# that process' threads.  For GPUs we use the same structure, but GPU processes
# are currently all considered to be single-threaded.
#
# The respective launch method is expected to create processes on the set of
# cpus and gpus thus specified, (node_1, cores 0 and 4; node_2, cores 1 and 5).
# The other reserved cores are for the application to spawn threads on
# (`cpu_threads=2`).
#
# A scheduler MAY attach other information to the `slots` structure, with the
# intent to support the launch methods to enact the placement decition made by
# the scheduler.  In fact, a scheduler may use a completely different slot
# structure than above - but then is likely bound to a specific launch method
# which can interpret that structure.  A notable example is the BG/Q torus
# scheduler which will only work in combination with the dplace launch methods.
# `lm_info` is an opaque field which allows to communicate specific settings
# from the lrms to the launch method.
#
# FIXME: `lm_info` should be communicated to the LM instances in creation, not
#        as part of the slots.  Its constant anyway, as lm_info is set only
#        once during lrms startup.
#
# NOTE:  While the nodelist resources are listed as strings above, we in fact
#        use a list of integers, to simplify some operations, and to
#        specifically avoid string copies on manipulations.  We only convert
#        to a stringlist for visual representation (`self.slot_status()`).
#
# NOTE:  The scheduler will allocate one core per node and GPU, as some startup
#        methods only allow process placements to *cores*, even if GPUs are
#        present and requested (hi aprun).  We should make this decision
#        dependent on the `lm_info` field - but at this point we don't have this
#        information (at least not readily available).
#
# TODO:  use named tuples for the slot structure to make the code more readable,
#        specifically for the LMs.
#

# NOTE:  The set of profiler events generated by this component are:
#
#        schedule_try    : search for unit resources starts    (uid: uid)
#        schedule_fail   : search for unit resources failed    (uid: uid)
#        schedule_ok     : search for unit resources succeeded (uid: uid)
#        unschedule_start: unit resource freeing starts        (uid: uid)
#        unschedule_stop : unit resource freeing stops         (uid: uid)
#
#        See also:
#        https://github.com/radical-cybertools/radical.pilot/blob/feature/ \
#                           events/docs/source/events.md \
#                           #agentschedulingcomponent-component
#
# ==============================================================================
#
class AgentSchedulingComponent(rpu.Component):

    # --------------------------------------------------------------------------
    #
    # the deriving schedulers should in general have the following structure in
    # self.odes:
    #
    #   self.nodes = [
    #     { 'name'  : 'name-of-node',
    #       'uid'   : 'uid-of-node',
    #       'cores' : '###---##-##-----',  # 16 cores, free/busy markers
    #       'gpus'  : '--',                #  2 GPUs,  free/busy markers
    #     }, ...
    #   ]
    #
    # The free/busy markers are defined in rp.constants.py, and are `-` and `#`,
    # respectively.  Some schedulers may need a more elaborate structures - but
    # where the above is suitable, it should be used for code consistency.
    #
    def __init__(self, cfg, session):

        self.nodes = None
        self._lrms = None
        self._uid  = ru.generate_id(cfg['owner'] + '.scheduling.%(counter)s',
                                    ru.ID_CUSTOM)

        self._uniform_waitpool = True   # TODO: move to cfg

        rpu.Component.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    # Once the component process is spawned, `initialize_child()` will be called
    # before control is given to the component's main loop.
    #
    def initialize_child(self):

        # register unit input channels
        self.register_input(rps.AGENT_SCHEDULING_PENDING,
                            rpc.AGENT_SCHEDULING_QUEUE, self._schedule_units)

        # register unit output channels
        self.register_output(rps.AGENT_EXECUTING_PENDING,
                             rpc.AGENT_EXECUTING_QUEUE)

        # we need unschedule updates to learn about units for which to free the
        # allocated cores.  Those updates MUST be issued after execution, ie.
        # by the AgentExecutionComponent.
        self.register_subscriber(rpc.AGENT_UNSCHEDULE_PUBSUB, self.unschedule_cb)

        # we don't want the unschedule above to compete with actual
        # scheduling attempts, so we move the re-scheduling of units from the
        # wait pool into a separate thread (ie. register a separate callback).
        # This is triggered by the unscheduled_cb.
        #
        # NOTE: we could use a local queue here.  Using a zmq bridge goes toward
        #       an distributed scheduler, and is also easier to implement right
        #       now, since `Component` provides the right mechanisms...
        self.register_publisher (rpc.AGENT_SCHEDULE_PUBSUB)
        self.register_subscriber(rpc.AGENT_SCHEDULE_PUBSUB, self.schedule_cb)

        # The scheduler needs the LRMS information which have been collected
        # during agent startup.  We dig them out of the config at this point.
        #
        # NOTE: this information is insufficient for the torus scheduler!
        self._pilot_id = self._cfg['pilot_id']
        self._lrms_info           = self._cfg['lrms_info']
        self._lrms_lm_info        = self._cfg['lrms_info']['lm_info']
        self._lrms_node_list      = self._cfg['lrms_info']['node_list']
        self._lrms_cores_per_node = self._cfg['lrms_info']['cores_per_node']
        self._lrms_gpus_per_node  = self._cfg['lrms_info']['gpus_per_node']

        if not self._lrms_node_list:
            raise RuntimeError("LRMS %s didn't _configure node_list."
                              % self._lrms_info['name'])

        if self._lrms_cores_per_node is None:
            raise RuntimeError("LRMS %s didn't _configure cores_per_node."
                              % self._lrms_info['name'])

        if self._lrms_gpus_per_node is None:
            raise RuntimeError("LRMS %s didn't _configure gpus_per_node."
                              % self._lrms_info['name'])

        # create and initialize the wait pool
        self._wait_pool = list()             # pool of waiting units
        self._wait_lock = threading.RLock()  # look on the above pool
        self._slot_lock = threading.RLock()  # lock slot allocation/deallocation

        # initialize the node list to be used by the scheduler.  A scheduler
        # instance may decide to overwrite or extend this structure.
        self.nodes = []
        for node, node_uid in self._lrms_node_list:
            self.nodes.append({
                'name' : node,
                'uid'  : node_uid,
                'cores': [rpc.FREE] * self._lrms_cores_per_node,
                'gpus' : [rpc.FREE] * self._lrms_gpus_per_node
            })

        # configure the scheduler instance
        self._configure()
        self._log.debug("slot status after  init      : %s", 
                        self.slot_status())


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate instance for the scheduler.
    #
    @classmethod
    def create(cls, cfg, session):

        # make sure that we are the base-class!
        if cls != AgentSchedulingComponent:
            raise TypeError("Scheduler Factory only available to base class!")

        name = cfg['scheduler']

        from .continuous_fifo import ContinuousFifo
        from .continuous      import Continuous
        from .scattered       import Scattered
        from .hombre          import Hombre
        from .torus           import Torus
        from .yarn            import Yarn
        from .spark           import Spark

        try:
            impl = {
                SCHEDULER_NAME_CONTINUOUS_FIFO : ContinuousFifo,
                SCHEDULER_NAME_CONTINUOUS      : Continuous,
                SCHEDULER_NAME_SCATTERED       : Scattered,
                SCHEDULER_NAME_HOMBRE          : Hombre,
                SCHEDULER_NAME_TORUS           : Torus,
                SCHEDULER_NAME_YARN            : Yarn,
                SCHEDULER_NAME_SPARK           : Spark
            }[name]

            impl = impl(cfg, session)
            return impl

        except KeyError:
            raise ValueError("Scheduler '%s' unknown or defunct" % name)


    # --------------------------------------------------------------------------
    #
    # Change the reserved state of slots (rpc.FREE or rpc.BUSY)
    #
    # NOTE: any scheduler implementation which uses a different nodelist
    #       structure MUST overload this method.
    #
    def _change_slot_states(self, slots, new_state):
        '''
        This function is used to update the state for a list of slots that
        have been allocated or deallocated.  For details on the data structure,
        see top of `base.py`.
        '''

        for node_name, node_uid, cores, gpus in slots['nodes']:

            # Find the entry in the the slots list

            # TODO: [Optimization] Assuming 'uid' is the ID of the node, it
            #       seems a bit wasteful to have to look at all of the nodes
            #       available for use if at most one node can have that uid.
            #       Maybe it would be worthwhile to simply keep a list of nodes
            #       that we would read, and keep a dictionary that maps the uid
            #       of the node to the location on the list?

            node = (n for n in self.nodes if n['uid'] == node_uid).next()
            assert(node)

            # iterate over cores/gpus in the slot, and update state
            for cslot in cores:
                for core in cslot:
                    node['cores'][core] = new_state

            for gslot in gpus:
                for gpu in gslot:
                    node['gpus'][gpu] = new_state


    # --------------------------------------------------------------------------
    #
    # NOTE: any scheduler implementation which uses a different nodelist
    #       structure MUST overload this method.
    def slot_status(self):
        '''
        Returns a multi-line string corresponding to the status of the node list
        '''

        ret = "|"
        for node in self.nodes:
            for core in node['cores']:
                if core == rpc.FREE  : ret += '-'
                else                 : ret += '#'
            ret += ':'
            for gpu in node['gpus']  :
                if gpu == rpc.FREE   : ret += '-'
                else                 : ret += '#'
            ret += '|'

        return ret


    # --------------------------------------------------------------------------
    #
    def _configure(self):
        raise NotImplementedError("_configure() missing for '%s'" % self.uid)


    # --------------------------------------------------------------------------
    #
    def _allocate_slot(self, cud):
        raise NotImplementedError("_allocate_slot() missing for '%s'" % self.uid)


    # --------------------------------------------------------------------------
    #
    def _release_slot(self, slots):
        raise NotImplementedError("_release_slot() missing for '%s'" % self.uid)


    # --------------------------------------------------------------------------
    #
    def _schedule_units(self, units):
        '''
        This is the main callback of the component, which is calledfor any
        incoming (set of) unit(s).  Units arriving here must always be in
        `AGENT_SCHEDULING_PENDING` state, and must always leave in either
        `AGENT_EXECUTING_PENDING` or in a FINAL state (`FAILED` or `CANCELED`).
        While handled by this method, the units will be in `AGENT_SCHEDULING`
        state.
        '''

        # unify handling of bulks / non-bulks
        if not isinstance(units, list):
            units = [units]

        # advance state, publish state change, do not push unit out.
        self.advance(units, rps.AGENT_SCHEDULING, publish=True, push=False)

        for unit in units:

            # we got a new unit to schedule.  Either we can place it
            # straight away and move it to execution, or we have to
            # put it in the wait pool.
            if self._try_allocation(unit):
                # we could schedule the unit - advance its state, notify worls
                # about the state change, and push the unit out toward the next
                # component.
                self.advance(unit, rps.AGENT_EXECUTING_PENDING, 
                             publish=True, push=True)
            else:
                # no resources available, put in wait queue
                with self._wait_lock :
                    self._wait_pool.append(unit)


    # --------------------------------------------------------------------------
    #
    def _try_allocation(self, unit):
        """
        attempt to allocate cores/gpus for a specific unit.
        """

        # needs to be locked as we try to acquire slots here, but slots are
        # freed in a different thread.  But we keep the lock duration short...
        with self._slot_lock :

            self._prof.prof('schedule_try', uid=unit['uid'])
            unit['slots'] = self._allocate_slot(unit['description'])

        # the lock is freed here
        if not unit['slots']:

            # signal the unit remains unhandled (Fales signals that failure)
            self._prof.prof('schedule_fail', uid=unit['uid'])
            return False


        # got an allocation, we can go off and launch the process
        self._prof.prof('schedule_ok', uid=unit['uid'])

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug("after  allocate   %s: %s", unit['uid'],
                            self.slot_status())
            self._log.debug("%s [%s/%s] : %s", unit['uid'],
                            unit['description']['cpu_processes'],
                            unit['description']['gpu_processes'],
                            pprint.pformat(unit['slots']))

        # True signals success
        return True


    # --------------------------------------------------------------------------
    #
    def _get_node_maps(self, cores, gpus, threads_per_proc):
        '''
        For a given set of cores and gpus, chunk them into sub-sets so that each
        sub-set can host one application process and all threads of that
        process.  Note that we currently consider all GPU applications to be
        single-threaded.

        example:
            cores  : [1, 2, 3, 4, 5, 6, 7, 8]
            gpus   : [1, 2]
            tpp    : 4
            result : [[1, 2, 3, 4], [5, 6, 7, 8]], [[1], [2]]

        For more details, see top level comment of `base.py`.
        '''

        core_map = list()
        gpu_map  = list()

        # make sure the core sets can host the requested number of threads
        assert(not len(cores) % threads_per_proc)
        n_procs =  len(cores) / threads_per_proc

        idx = 0
        for p in range(n_procs):
            p_map = list()
            for t in range(threads_per_proc):
                p_map.append(cores[idx])
                idx += 1
            core_map.append(p_map)

        if idx != len(cores):
            self._log.debug('%s -- %s -- %s -- %s',
                            idx, len(cores), cores, n_procs)
        assert(idx == len(cores))

        # gpu procs are considered single threaded right now (FIXME)
        for g in gpus:
            gpu_map.append([g])

        return core_map, gpu_map


    # --------------------------------------------------------------------------
    #
    def unschedule_cb(self, topic, msg):
        """
        release (for whatever reason) all slots allocated to this unit
        """

        unit = msg

        if not unit['slots']:
            # Nothing to do -- how come?
            self._log.error("cannot unschedule: %s (no slots)" % unit)
            return True

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug("before unschedule %s: %s", unit['uid'], 
                            self.slot_status())

        # needs to be locked as we try to release slots, but slots are acquired
        # in a different thread....
        with self._slot_lock :
            self._prof.prof('unschedule_start', uid=unit['uid'])
            self._release_slot(unit['slots'])
            self._prof.prof('unschedule_stop',  uid=unit['uid'])

        # notify the scheduling thread, ie. trigger an attempt to use the freed
        # slots for units waiting in the wait pool.
        self.publish(rpc.AGENT_SCHEDULE_PUBSUB, unit)

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug("after  unschedule %s: %s", unit['uid'], 
                            self.slot_status())

        # return True to keep the cb registered
        return True


    # --------------------------------------------------------------------------
    #
    def schedule_cb(self, topic, msg):
        '''
        This cb is triggered after a unit's resources became available again, so
        we can attempt to schedule units from the wait pool.
        '''

        # we ignore any passed unit.  In principle the unit info could be used to
        # determine which slots have been freed.  No need for that optimization
        # right now.  This will become interesting once schedule becomes too
        # expensive.
        # FIXME: optimization

        unit = msg

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug("before schedule   %s: %s", unit['uid'],
                            self.slot_status())

        # cycle through wait queue, and see if we get anything placed now.  We
        # cycle over a copy of the list, so that we can modify the list on the
        # fly,without locking the whole loop.  However, this is costly, too.
        for unit in self._wait_pool[:]:

            if self._try_allocation(unit):

                # allocated unit -- advance it
                self.advance(unit, rps.AGENT_EXECUTING_PENDING, publish=True, push=True)

                # remove it from the wait queue
                with self._wait_lock :
                    self._wait_pool.remove(unit)

            else:
                # Break out of this loop if we didn't manage to schedule a task
                # FIXME: this assumes that no smaller or otherwise more suitable
                #        CUs come after this one - which is naive, ie. wrong.
                # NOTE:  This assumption does indeed break for the fifo
                #        scheduler, so we disable this now for non-uniform cases
                if self._uniform_waitpool:
                    break

        # return True to keep the cb registered
        return True


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys

verbose  = os.environ.get('RADICAL_PILOT_VERBOSE', 'REPORT')
os.environ['RADICAL_PILOT_VERBOSE'] = verbose

import radical.pilot as rp
import radical.utils as ru


# ------------------------------------------------------------------------------
#
# READ the RADICAL-Pilot documentation: http://radicalpilot.readthedocs.org/
#
# ------------------------------------------------------------------------------


#------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # we use a reporter class for nicer output
    report = ru.Reporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    # use the resource specified as argument, fall back to localhost
    if   len(sys.argv)  > 2: report.exit('Usage:\t%s [resource]\n\n' % sys.argv[0])
    elif len(sys.argv) == 2: resource = sys.argv[1]
    else                   : resource = 'local.localhost'

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # read the config used for resource details
        report.info('read config')
        config = ru.read_json('%s/config.json' % os.path.dirname(os.path.abspath(__file__)))
        report.ok('>>ok\n')

        report.header('submit pilots')

        # Add a Pilot Manager. Pilot managers manage one or more ComputePilots.
        pmgr = rp.PilotManager(session=session)

        # Define an [n]-core local pilot that runs for [x] minutes
        # Here we use a dict to initialize the description object
        pd_init = {
                'resource'      : resource,
                'runtime'       : 15,  # pilot runtime (min)
                'exit_on_error' : True,
                'project'       : config[resource]['project'],
                'queue'         : config[resource]['queue'],
                'access_schema' : config[resource]['schema'],
                'cores'         : config[resource]['cores'],
                }
        pdesc = rp.ComputePilotDescription(pd_init)

        # Launch the pilot.
        pilot = pmgr.submit_pilots(pdesc)

        # Create a workload of char-counting a simple file.  We first create the
        # file right here, and stage it to the pilot 'shared_data' space
        os.system('hostname >  input.dat')
        os.system('date     >> input.dat')

        # Synchronously stage the data to the pilot
        report.info('stage shared data')
        pilot.stage_in({'source': 'client:///input.dat',
                        'target': 'pilot:///input.dat',
                        'action': rp.TRANSFER})
        report.ok('>>ok\n')


        report.header('submit units')

        # Register the ComputePilot in a UnitManager object.
        umgr = rp.UnitManager(session=session)
        umgr.add_pilots(pilot)

        n = 128   # number of units to run
        report.info('create %d unit description(s)\n\t' % n)

        cuds = list()
        for i in range(0, n):

            # create a new CU description, and fill it.
            # Here we don't use dict initialization.
            cud = rp.ComputeUnitDescription()
            cud.executable     = '/bin/echo'
            cud.arguments      = ['-c', 'input.dat', '%d' % i]
            cud.input_staging  = {'source': 'pilot:///input.dat', 
                                  'target': 'unit:///input.dat',
                                  'action': rp.LINK}
            cud.output_staging = {'source': 'unit:///STDOUT',
                                  'target': 'pilot:///STDOUT.%06d'%i,
                                  'action': rp.COPY }
            cuds.append(cud)
            report.progress()
        report.ok('>>ok\n')

        # Submit the previously created ComputeUnit descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning ComputeUnits to the ComputePilots.
        units = umgr.submit_units(cuds)

        # Wait for all compute units to reach a final state (DONE, CANCELED or FAILED).
        report.header('gather results')
        umgr.wait_units()
    
        report.info('\n')
        for unit in units:
            report.plain('  * %s: %s, exit: %3s, out: %s\n' \
                    % (unit.uid, unit.state[:4], 
                        unit.exit_code, unit.stdout.strip()[:35]))
    
        # delete the sample input files
        os.system('rm input.dat')


  # except Exception as e:
  #     # Something unexpected happened in the pilot code above
  #     report.error('caught Exception: %s\n' % e)
  #     raise 
  #
  # except (KeyboardInterrupt, SystemExit) as e:
  #     # the callback called sys.exit(), and we can here catch the
  #     # corresponding KeyboardInterrupt exception for shutdown.  We also catch
  #     # SystemExit (which gets raised if the main threads exits for some other
  #     # reason).
  #     report.warn('exit requested\n')

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.  This will kill all remaining pilots.
        report.header('finalize')
        session.close()

    report.header()


#-------------------------------------------------------------------------------





import os
import sys
import zmq
import copy
import math
import time
import errno
import pprint
import msgpack

import Queue           as pyq
import setproctitle    as spt
import threading       as mt
import multiprocessing as mp

import radical.utils   as ru

from .misc import hostip as rpu_hostip


# --------------------------------------------------------------------------
# defines for queue roles
#
QUEUE_INPUT   = 'input'
QUEUE_BRIDGE  = 'bridge'
QUEUE_OUTPUT  = 'output'
QUEUE_ROLES   = [QUEUE_INPUT, QUEUE_BRIDGE, QUEUE_OUTPUT]

_BRIDGE_TIMEOUT  =     1  # how long to wait for bridge startup
_LINGER_TIMEOUT  =   250  # ms to linger after close
_HIGH_WATER_MARK =     0  # number of messages to buffer before dropping


# --------------------------------------------------------------------------
#
# zmq will (rightly) barf at interrupted system calls.  We are able to rerun
# those calls.
#
# FIXME: how does that behave wrt. tomeouts?  We probably should include
#        an explicit timeout parameter.
#
# kudos: https://gist.github.com/minrk/5258909
#
def _uninterruptible(f, *args, **kwargs):
    cnt = 0
    while True:
        cnt += 1
        try:
            return f(*args, **kwargs)
        except zmq.ContextTerminated as e:
            return None
        except zmq.ZMQError as e:
            if e.errno == errno.EINTR:
                if cnt > 10:
                    raise
                # interrupted, try again
                continue
            else:
                # real error, raise it
                raise


# ------------------------------------------------------------------------------
#
# Communication between components is done via queues.  The semantics we expect
# (and which is what is matched by the native Python Queue.Queue), is:
#
#   - multiple upstream   components put messages onto the same queue (input)
#   - multiple downstream components get messages from the same queue (output)
#   - local order of messages is maintained
#   - message routing is fair: whatever downstream component calls 'get' first
#     will get the next message (bridge)
#
# Additionally, we require Queues to be uni-directional, ie. Queues have an
# in-end for which one can call 'put()', and and out-end, for which one can call
# 'get()'.  We implement the interface of Queue.Queue:
#
#   put(msg)
#   get()
#   get_nowait()
#
# Not implemented is, at the moment:
#
#   qsize
#   empty
#   full 
#   put(msg, block, timeout)
#   put_nowait
#   get(block, timeout)
#   task_done
#
# Our Queue additionally takes 'name', 'role' and 'address' parameter on the
# constructor.  'role' can be 'input', 'bridge' or 'output', where 'input' is
# the end of a queue one can 'put()' messages into, and 'output' the end of the
# queue where one can 'get()' messages from. A 'bridge' acts as as a message
# forwarder.  'address' denominates a connection endpoint, and 'name' is
# a unique identifier: if multiple instances in the current process space use
# the same identifier, they will get the same queue instance.


# ==============================================================================
#
class Queue(ru.Process):

    def __init__(self, session, qname, role, cfg, addr=None):
        """
        This Queue type sets up an zmq channel of this kind:

            input \            / output
                   -- bridge -- 
            input /            \ output

        ie. any number of inputs can 'zmq.push()' to a bridge (which
        'zmq.pull()'s), and any number of outputs can 'zmq.request()' 
        messages from the bridge (which 'zmq.response()'s).

        The bridge is the entity which 'bind()'s network interfaces, both input
        and output type endpoints 'connect()' to it.  It is the callees
        responsibility to ensure that only one bridge of a given type exists.

        Addresses are of the form 'tcp://host:port'.  Both 'host' and 'port' can
        be wildcards for BRIDGE roles -- the bridge will report the in and out
        addresses as obj.addr_in and obj.addr_out.  """

        self._session = session
        self._qname   = qname
        self._role    = role
        self._cfg     = copy.deepcopy(cfg)
        self._addr    = addr

        assert(self._role in QUEUE_ROLES), 'invalid role %s' % self._role

        self._uid = "%s.%s" % (self._qname.replace('_', '.'), self._role)
        self._uid = ru.generate_id(self._uid)
        self._log = self._session._get_logger(name=self._uid, 
                         level=self._cfg.get('log_level'))

        super(Queue, self).__init__(name=self._uid, log=self._log)

        # avoid superfluous logging calls in critical code sections
        if self._log.getEffectiveLevel() == 10: # logging.DEBUG:
            self._debug = True
        else:
            self._debug = False

        self._lock       = mt.RLock()     # for _requested
        self._requested  = False          # send/recv sync
        self._addr_in    = None           # bridge input  addr
        self._addr_out   = None           # bridge output addr
        self._stall_hwm  = cfg.get('stall_hwm', 1)
        self._bulk_size  = cfg.get('bulk_size', 1)

        if not self._addr:
            self._addr = 'tcp://*:*'

        self._log.info("create %s - %s - %s", self._qname, self._role, self._addr)

        self._q    = None           # the zmq queue
        self._in   = None
        self._out  = None
        self._ctx  = None


        # ----------------------------------------------------------------------
        # behavior depends on the role...
        if self._role == QUEUE_INPUT:

            self._ctx = zmq.Context()
            self._session._to_destroy.append(self._ctx)

            self._q   = self._ctx.socket(zmq.PUSH)
            self._q.linger = _LINGER_TIMEOUT
            self._q.hwm    = _HIGH_WATER_MARK
            self._q.connect(self._addr)
            self.start(spawn=False)


        # ----------------------------------------------------------------------
        elif self._role == QUEUE_BRIDGE:

            # we expect bridges to always use a port wildcard. Make sure
            # that's the case
            elems = self._addr.split(':')
            if len(elems) > 2 and elems[2] and elems[2] != '*':
                raise RuntimeError('wildcard port (*) required for bridge addresses (%s)' \
                                % self._addr)

            self._pqueue = mp.Queue()
            self.start()

            try:
                [addr_in, addr_out] = self._pqueue.get(True, _BRIDGE_TIMEOUT)

                # store addresses
                self._addr_in  = ru.Url(addr_in)
                self._addr_out = ru.Url(addr_out)

                # use the local hostip for bridge addresses
                self._addr_in.host  = rpu_hostip()
                self._addr_out.host = rpu_hostip()

            except pyq.Empty as e:
                raise RuntimeError ("bridge did not come up! (%s)" % e)


        # ----------------------------------------------------------------------
        elif self._role == QUEUE_OUTPUT:

            self._ctx = zmq.Context()
            self._session._to_destroy.append(self._ctx)

            self._q   = self._ctx.socket(zmq.REQ)
            self._q.linger = _LINGER_TIMEOUT
            self._q.hwm    = _HIGH_WATER_MARK
            self._q.connect(self._addr)
            self.start(spawn=False)


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def qname(self):
        return self._qname

    @property
    def role(self):
        return self._role

    @property
    def addr(self):
        return self._addr

    @property
    def addr_in(self):
        assert(self._role == QUEUE_BRIDGE), 'addr_in only set on bridges'
        return self._addr_in

    @property
    def addr_out(self):
        assert(self._role == QUEUE_BRIDGE), 'addr_out only set on bridges'
        return self._addr_out


    # --------------------------------------------------------------------------
    # 
    def ru_initialize_child(self):

        assert(self._role == QUEUE_BRIDGE), 'only bridges can be started'

        self._uid = self._uid + '.child'
        self._log = self._session._get_logger(name=self._uid, 
                         level=self._cfg.get('log_level'))

        spt.setproctitle('rp.%s' % self._uid)
        self._log.info('start bridge %s on %s', self._uid, self._addr)

        # FIXME: should we cache messages coming in at the pull/push 
        #        side, so as not to block the push end?

        self._ctx = zmq.Context()
        self._session._to_destroy.append(self._ctx)

        self._in = self._ctx.socket(zmq.PULL)
        self._in.linger = _LINGER_TIMEOUT
        self._in.hwm    = _HIGH_WATER_MARK
        self._in.bind(self._addr)

        self._out = self._ctx.socket(zmq.REP)
        self._out.linger = _LINGER_TIMEOUT
        self._out.hwm    = _HIGH_WATER_MARK
        self._out.bind(self._addr)

        # communicate the bridge ports to the parent process
        _addr_in  = self._in.getsockopt( zmq.LAST_ENDPOINT)
        _addr_out = self._out.getsockopt(zmq.LAST_ENDPOINT)

        self._pqueue.put([_addr_in, _addr_out])

        self._log.info('bound bridge %s to %s : %s', self._uid, _addr_in, _addr_out)

        # start polling for messages
        self._poll = zmq.Poller()
        self._poll.register(self._out, zmq.POLLIN)


    # --------------------------------------------------------------------------
    # 
    def ru_finalize_common(self):

        if self._q   : self._q   .close()
        if self._in  : self._in  .close()
        if self._out : self._out .close()
        if self._ctx : self._ctx.destroy()


    # --------------------------------------------------------------------------
    # 
    def work_cb(self):

        # We wait for an incoming message.  When one is received,
        # we'll poll all outgoing sockets for requests, and the
        # forward the message to whoever requested it.
        #
        # If so configured, we can stall messages until reaching
        # a certain high-water-mark, and upon reaching that will
        # release all messages at once.  When stalling for such set
        # of messages, we wait for self._stall_hwm messages, and
        # then forward those to whatever output channel requesting
        # them (individually),
        # FIXME: we may want to release all stalled messages after
        #        some (configurable) timeout?
        # FIXME: is it worth introducing a 'flush' command or
        #        message type?
        # NOTE:  hwm collection interferes with termination:
        #        while we wait for messages to arrive, we will not leave this
        #        work_cb, and thus will not allow the `ru.Process` main loop to
        #        terminate gracefully as needed.  One solution would be to
        #        inject manual statements to check for termination conditions.
        #        The other option would be to collect the messages across
        #        multiple work_cb invocations -- but the latter options would
        #        significantly increase the code complexity, as we would need to
        #        maintain state across method invocations, so we opt for the
        #        first option.
        hwm  = self._stall_hwm
        bulk = self._bulk_size

        msgs = list()
        while len(msgs) < hwm:
            data = None
            while self.is_alive(strict=False):
                if _uninterruptible(self._in.poll, flags=zmq.POLLIN, 
                                   timeout=1000):
                    data = _uninterruptible(self._in.recv)
                    break
            if not data:
                if not self.is_alive(strict=False):
                    self._log.warn('not alive anymore?')
                    return False
            msg = msgpack.unpackb(data) 
            if isinstance(msg, list): 
                msgs += msg
            else: 
                msgs.append(msg)
            self._log.debug('stall %s/%s', len(msgs), hwm)
        self._log.debug('hwm   %s/%s', len(msgs), hwm)

        # if 'bulk' is '0', we send all messages as
        # a single bulk.  Otherwise, we chop them up
        # into bulks of the given size
        if bulk <= 0:
            nbulks = 1
            bulks  = [msgs]
        else:
            nbulks = int(math.ceil(len(msgs) / float(bulk)))
            bulks  = ru.partition(msgs, nbulks)

        while bulks:
            # timeout in ms
            events = dict(_uninterruptible(self._poll.poll, 1000))

            if self._out in events:

                req  = _uninterruptible(self._out.recv)
                data = msgpack.packb(bulks.pop(0)) 
                _uninterruptible(self._out.send, data)

                # go to next message/bulk (break while loop)
                self._log.debug('sent  %s [hwm: %s]', (nbulks-len(bulks)), hwm)

        return True


    # --------------------------------------------------------------------------
    #
    def put(self, msg):

        if not self._role == QUEUE_INPUT:
            raise RuntimeError("queue %s (%s) can't put()" % (self._qname, self._role))

      # if self._debug:
      #     self._log.debug("-> %s", pprint.pformat(msg))
        data = msgpack.packb(msg) 
        _uninterruptible(self._q.send, data)


    # --------------------------------------------------------------------------
    #
    def get(self):

        if not self._role == QUEUE_OUTPUT:
            raise RuntimeError("queue %s (%s) can't get()" % (self._qname, self._role))

        _uninterruptible(self._q.send, 'request')

        data = _uninterruptible(self._q.recv)
        msg  = msgpack.unpackb(data) 
      # if self._debug:
      #     self._log.debug("<- %s", pprint.pformat(msg))
        return msg


    # --------------------------------------------------------------------------
    #
    def get_nowait(self, timeout=None): # timeout in ms

        if not self._role == QUEUE_OUTPUT:
            raise RuntimeError("queue %s (%s) can't get_nowait()" % (self._qname, self._role))

        with self._lock: # need to protect self._requested

            if not self._requested:
                # we can only send the request once per recieval
                _uninterruptible(self._q.send, 'request')
                self._requested = True

          # try:
          #     msg = self._q.recv_json(flags=zmq.NOBLOCK)
          #     self._requested = False
          #     if self._debug:
          #         self._log.debug("<< %s", pprint.pformat(msg))
          #     return msg
          #
          # except zmq.Again:
          #     return None

            if _uninterruptible(self._q.poll, flags=zmq.POLLIN, timeout=timeout):
                data = _uninterruptible(self._q.recv)
                msg  = msgpack.unpackb(data) 
                self._requested = False
              # if self._debug:
              #     self._log.debug("<< %s", pprint.pformat(msg))
                return msg

            else:
                return None


# ------------------------------------------------------------------------------

